#include "CloudComm.h"
#include "TimingSingleton.h"
#include "SecureRandom.h"
#include "IoTString.h"
#include "Error.h"
#include "URL.h"
#include "Mac.h"
#include "Table.h"
#include "Slot.h"
#include "Crypto.h"
#include "ByteBuffer.h"
#include <sys/socket.h>
#include <unistd.h>

/**
 * Empty Constructor needed for child class.
 */
CloudComm::CloudComm() :
	baseurl(NULL),
	key(NULL),
	mac(NULL),
	password(NULL),
	random(NULL),
	salt(NULL),
	table(NULL),
	listeningPort(-1),
	localServerThread(NULL),
	doEnd(false),
	timer(TimingSingleton_getInstance()),
	getslot(new Array<char>("getslot", 7)),
	putslot(new Array<char>("putslot", 7))
{
}

void *threadWrapper(void *cloud) {
	CloudComm *c = (CloudComm *) cloud;
	c->localServerWorkerFunction();
	return NULL;
}

/**
 * Constructor for actual use. Takes in the url and password.
 */
CloudComm::CloudComm(Table *_table,  IoTString *_baseurl, IoTString *_password, int _listeningPort) :
	baseurl(_baseurl),
	key(NULL),
	mac(NULL),
	password(_password),
	random(new SecureRandom()),
	salt(NULL),
	table(_table),
	listeningPort(_listeningPort),
	localServerThread(NULL),
	doEnd(false),
	timer(TimingSingleton_getInstance()) {
	if (listeningPort > 0) {
		pthread_create(&localServerThread, NULL, threadWrapper, this);
	}
}

/**
 * Generates Key from password.
 */
AESKey *CloudComm::initKey() {
	try {
		AESKey *key = new AESKey(password->internalBytes(),
														 salt,
														 65536,
														 128);
		return key;
	} catch (Exception *e) {
		throw new Error("Failed generating key.");
	}
}

/**
 * Inits all the security stuff
 */

void CloudComm::initSecurity() {
	// try to get the salt and if one does not exist set one
	if (!getSalt()) {
		//Set the salt
		setSalt();
	}

	initCrypt();
}

/**
 * Inits the HMAC generator.
 */
void CloudComm::initCrypt() {
	if (password == NULL) {
		return;
	}
	try {
		key = initKey();
		password = NULL;// drop password
		mac = new Mac();
		mac->init(key);
	} catch (Exception *e) {
		throw new Error("Failed To Initialize Ciphers");
	}
}

/*
 * Builds the URL for the given request.
 */
IoTString *CloudComm::buildRequest(bool isput, int64_t sequencenumber, int64_t maxentries) {
	const char *reqstring = isput ? "req=putslot" : "req=getslot";
	char *buffer = (char *) malloc(baseurl->length() + 200);
	memcpy(buffer, baseurl->internalBytes(), baseurl->length());
	int offset = baseurl->length();
	offset += sprintf(&buffer[offset], "?%s&seq=%" PRId64, reqstring, sequencenumber);
	if (maxentries != 0)
		sprintf(&buffer[offset], "&max=%" PRId64, maxentries);
	IoTString *urlstr = new IoTString(buffer);
	return urlstr;
}

int openURL(IoTString *url, bool isPost) {
	return 0;
}

int createSocket(IoTString *host, int port) {
	return 0;
}

int createSocket(int port) {
	return 0;
}

int acceptSocket(int socket) {
	return 0;
}

void writeSocketData(int fd, Array<char> *data) {}

void writeSocketInt(int fd, int value) {}

int readSocketInt(int fd) {return 0;}

void readSocketData(int fd, Array<char> *data) {}

void writeURLData(int fd, Array<char> *data) {
}

void readURLData(int fd, Array<char> *output) {
}

int readURLInt(int fd) {
	return 0;
}

int getResponseCode(int fd) {
	return 0;
}

void CloudComm::setSalt() {
	if (salt != NULL) {
		// Salt already sent to server so don't set it again
		return;
	}

	int fd = -1;
	try {
		Array<char> *saltTmp = new Array<char>(CloudComm_SALT_SIZE);
		random->nextBytes(saltTmp);

		char *buffer = (char *) malloc(baseurl->length() + 100);
		memcpy(buffer, baseurl->internalBytes(), baseurl->length());
		int offset = baseurl->length();
		offset += sprintf(&buffer[offset], "?req=setsalt");
		IoTString *urlstr = new IoTString(buffer);
		free(buffer);

		timer->startTime();
		fd = openURL(urlstr, true);
		writeURLData(fd, saltTmp);

		int responsecode = getResponseCode(fd);
		if (responsecode != HttpURLConnection_HTTP_OK) {
			throw new Error("Invalid response");
		}

		timer->endTime();
		salt = saltTmp;
	} catch (Exception *e) {
		timer->endTime();
		throw new ServerException("Failed setting salt", ServerException_TypeConnectTimeout);
	}
}

bool CloudComm::getSalt() {
	int fd = -1;
	IoTString *urlstr = NULL;

	try {
		char *buffer = (char *) malloc(baseurl->length() + 100);
		memcpy(buffer, baseurl->internalBytes(), baseurl->length());
		int offset = baseurl->length();
		offset += sprintf(&buffer[offset], "?req=getsalt");
		urlstr = new IoTString(buffer);
		free(buffer);
	} catch (Exception *e) {
		throw new Error("getSlot failed");
	}
	try {
		timer->startTime();
		fd = openURL(urlstr, true);
		timer->endTime();
	} catch (SocketTimeoutException *e) {
		timer->endTime();
		throw new ServerException("getSalt failed", ServerException_TypeConnectTimeout);
	} catch (Exception *e) {
		throw new Error("getSlot failed");
	}

	try {
		timer->startTime();
		int responsecode = getResponseCode(fd);
		if (responsecode != HttpURLConnection_HTTP_OK) {
			throw new Error("Invalid response");
		}
		int salt_length = readURLInt(fd);
		Array<char> *tmp = new Array<char>(salt_length);
		readURLData(fd, tmp);
		salt = tmp;
		timer->endTime();
		return true;
	} catch (SocketTimeoutException *e) {
		timer->endTime();
		throw new ServerException("getSalt failed", ServerException_TypeInputTimeout);
	} catch (Exception *e) {
		throw new Error("getSlot failed");
	}
}

Array<char> *CloudComm::createIV(int64_t machineId, int64_t localSequenceNumber) {
	ByteBuffer *buffer = ByteBuffer_allocate(CloudComm_IV_SIZE);
	buffer->putLong(machineId);
	int64_t localSequenceNumberShifted = localSequenceNumber << 16;
	buffer->putLong(localSequenceNumberShifted);
	return buffer->array();
}

Array<char> *AESEncrypt(Array<char> *ivBytes, AESKey *key, Array<char> *data) {
	return NULL;
}

Array<char> *AESDecrypt(Array<char> *ivBytes, AESKey *key, Array<char> *data) {
	return NULL;
}

Array<char> *CloudComm::encryptSlotAndPrependIV(Array<char> *rawData, Array<char> *ivBytes) {
	try {
		Array<char> *encryptedBytes = AESEncrypt(ivBytes, key, rawData);
		Array<char> *chars = new Array<char>(encryptedBytes->length() + CloudComm_IV_SIZE);
		System_arraycopy(ivBytes, 0, chars, 0, ivBytes->length());
		System_arraycopy(encryptedBytes, 0, chars, CloudComm_IV_SIZE, encryptedBytes->length());

		return chars;
	} catch (Exception *e) {
		throw new Error("Failed To Encrypt");
	}
}

Array<char> *CloudComm::stripIVAndDecryptSlot(Array<char> *rawData) {
	try {
		Array<char> *ivBytes = new Array<char>(CloudComm_IV_SIZE);
		Array<char> *encryptedBytes = new Array<char>(rawData->length() - CloudComm_IV_SIZE);
		System_arraycopy(rawData, 0, ivBytes, 0, CloudComm_IV_SIZE);
		System_arraycopy(rawData, CloudComm_IV_SIZE, encryptedBytes, 0, encryptedBytes->length());
		return AESDecrypt(ivBytes, key, encryptedBytes);
	} catch (Exception *e) {
		throw new Error("Failed To Decrypt");
	}
}

/*
 * API for putting a slot into the queue.  Returns NULL on success.
 * On failure, the server will send slots with newer sequence
 * numbers.
 */
Array<Slot *> *CloudComm::putSlot(Slot *slot, int max) {
	int fd = -1;
	try {
		if (salt == NULL) {
			if (!getSalt()) {
				throw new ServerException("putSlot failed", ServerException_TypeSalt);
			}
			initCrypt();
		}

		int64_t sequencenumber = slot->getSequenceNumber();
		Array<char> *slotBytes = slot->encode(mac);
		Array<char> *chars = encryptSlotAndPrependIV(slotBytes, slot->getSlotCryptIV());
		IoTString *url = buildRequest(true, sequencenumber, max);
		timer->startTime();
		fd = openURL(url, true);
		writeURLData(fd, chars);
		timer->endTime();
	} catch (ServerException *e) {
		timer->endTime();
		throw e;
	} catch (SocketTimeoutException *e) {
		timer->endTime();
		throw new ServerException("putSlot failed", ServerException_TypeConnectTimeout);
	} catch (Exception *e) {
		throw new Error("putSlot failed");
	}

	try {
		timer->startTime();
		Array<char> *resptype = new Array<char>(7);
		readURLData(fd, resptype);
		timer->endTime();

		if (resptype->equals(getslot)) {
			return processSlots(fd);
		} else if (resptype->equals(putslot)) {
			return NULL;
		} else
			throw new Error("Bad response to putslot");
	} catch (SocketTimeoutException *e) {
		timer->endTime();
		throw new ServerException("putSlot failed", ServerException_TypeInputTimeout);
	} catch (Exception *e) {
		throw new Error("putSlot failed");
	}
}

/**
 * Request the server to send all slots with the given
 * sequencenumber or newer->
 */
Array<Slot *> *CloudComm::getSlots(int64_t sequencenumber) {
	int fd = -1;
	try {
		if (salt == NULL) {
			if (!getSalt()) {
				throw new ServerException("getSlots failed", ServerException_TypeSalt);
			}
			initCrypt();
		}

		IoTString *url = buildRequest(false, sequencenumber, 0);
		timer->startTime();
		fd = openURL(url, true);
		timer->endTime();
	} catch (SocketTimeoutException *e) {
		timer->endTime();
		throw new ServerException("getSlots failed", ServerException_TypeConnectTimeout);
	} catch (ServerException *e) {
		timer->endTime();

		throw e;
	} catch (Exception *e) {
		throw new Error("getSlots failed");
	}

	try {
		timer->startTime();
		Array<char> *resptype = new Array<char>(7);
		readURLData(fd, resptype);
		timer->endTime();
		if (!resptype->equals(getslot))
			throw new Error("Bad Response: ");

		return processSlots(fd);
	} catch (SocketTimeoutException *e) {
		timer->endTime();
		throw new ServerException("getSlots failed", ServerException_TypeInputTimeout);
	} catch (Exception *e) {
		throw new Error("getSlots failed");
	}
}

/**
 * Method that actually handles building Slot objects from the
 * server response.  Shared by both putSlot and getSlots.
 */
Array<Slot *> *CloudComm::processSlots(int fd) {
	int numberofslots = readURLInt(fd);
	Array<int> *sizesofslots = new Array<int>(numberofslots);
	Array<Slot *> *slots = new Array<Slot *>(numberofslots);

	for (int i = 0; i < numberofslots; i++)
		sizesofslots->set(i, readURLInt(fd));
	for (int i = 0; i < numberofslots; i++) {
		Array<char> *rawData = new Array<char>(sizesofslots->get(i));
		readURLData(fd, rawData);
		Array<char> *data = stripIVAndDecryptSlot(rawData);
		slots->set(i, Slot_decode(table, data, mac));
	}
	return slots;
}

Array<char> *CloudComm::sendLocalData(Array<char> *sendData, int64_t localSequenceNumber, IoTString *host, int port) {
	if (salt == NULL)
		return NULL;
	try {
		printf("Passing Locally\n");
		mac->update(sendData, 0, sendData->length());
		Array<char> *genmac = mac->doFinal();
		Array<char> *totalData = new Array<char>(sendData->length() + genmac->length());
		System_arraycopy(sendData, 0, totalData, 0, sendData->length());
		System_arraycopy(genmac, 0, totalData, sendData->length(), genmac->length());

		// Encrypt the data for sending
		Array<char> *iv = createIV(table->getMachineId(), table->getLocalSequenceNumber());
		Array<char> *encryptedData = encryptSlotAndPrependIV(totalData, iv);

		// Open a TCP socket connection to a local device
		int socket = createSocket(host, port);

		timer->startTime();
		// Send data to output (length of data, the data)
		writeSocketInt(socket, encryptedData->length());
		writeSocketData(socket, encryptedData);

		int lengthOfReturnData = readSocketInt(socket);
		Array<char> *returnData = new Array<char>(lengthOfReturnData);
		readSocketData(socket, returnData);
		timer->endTime();
		returnData = stripIVAndDecryptSlot(returnData);

		// We are done with this socket
		close(socket);
		mac->update(returnData, 0, returnData->length() - CloudComm_HMAC_SIZE);
		Array<char> *realmac = mac->doFinal();
		Array<char> *recmac = new Array<char>(CloudComm_HMAC_SIZE);
		System_arraycopy(returnData, returnData->length() - realmac->length(), recmac, 0, realmac->length());

		if (!recmac->equals(realmac))
			throw new Error("Local Error: Invalid HMAC!  Potential Attack!");

		Array<char> *returnData2 = new Array<char>(lengthOfReturnData - recmac->length());
		System_arraycopy(returnData, 0, returnData2, 0, returnData2->length());

		return returnData2;
	} catch (Exception *e) {
		printf("Exception\n");
	}

	return NULL;
}

void CloudComm::localServerWorkerFunction() {
	int inputSocket = -1;

	try {
		// Local server socket
		inputSocket = createSocket(listeningPort);
	} catch (Exception *e) {
		throw new Error("Local server setup failure...");
	}

	while (!doEnd) {
		try {
			// Accept incoming socket
			int socket = acceptSocket(inputSocket);

			// Get the encrypted data from the server
			int dataSize = readSocketInt(socket);
			Array<char> *readData = new Array<char>(dataSize);
			readSocketData(socket, readData);
			timer->endTime();

			// Decrypt the data
			readData = stripIVAndDecryptSlot(readData);
			mac->update(readData, 0, readData->length() - CloudComm_HMAC_SIZE);
			Array<char> *genmac = mac->doFinal();
			Array<char> *recmac = new Array<char>(CloudComm_HMAC_SIZE);
			System_arraycopy(readData, readData->length() - recmac->length(), recmac, 0, recmac->length());

			if (!recmac->equals(genmac))
				throw new Error("Local Error: Invalid HMAC!  Potential Attack!");

			Array<char> *returnData = new Array<char>(readData->length() - recmac->length());
			System_arraycopy(readData, 0, returnData, 0, returnData->length());

			// Process the data
			Array<char> *sendData = table->acceptDataFromLocal(returnData);
			mac->update(sendData, 0, sendData->length());
			Array<char> *realmac = mac->doFinal();
			Array<char> *totalData = new Array<char>(sendData->length() + realmac->length());
			System_arraycopy(sendData, 0, totalData, 0, sendData->length());
			System_arraycopy(realmac, 0, totalData, sendData->length(), realmac->length());

			// Encrypt the data for sending
			Array<char> *iv = createIV(table->getMachineId(), table->getLocalSequenceNumber());
			Array<char> *encryptedData = encryptSlotAndPrependIV(totalData, iv);

			timer->startTime();
			// Send data to output (length of data, the data)
			writeSocketInt(socket, encryptedData->length());
			writeSocketData(socket, encryptedData);
			close(socket);
		} catch (Exception *e) {
		}
	}

	if (inputSocket != -1) {
		try {
			close(inputSocket);
		} catch (Exception *e) {
			throw new Error("Local server close failure...");
		}
	}
}

void CloudComm::closeCloud() {
	doEnd = true;

	if (localServerThread != NULL) {
		if (pthread_join(localServerThread, NULL) != 0)
			throw new Error("Local Server thread join issue...");
	}
}
